function p = kspaceLineRecon(p, dx, dt, c, varargin)
%KSPACELINERECON    2D linear FFT reconstruction.
%
% DESCRIPTION:
%       kspaceLineRecon takes an acoustic pressure time-series p_tx
%       recorded over an evenly spaced array of sensor points on a line,
%       and constructs an estimate of the initial acoustic pressure
%       distribution that gave rise to those measurements using an
%       algorithm based on the FFT. The input p_tx must be indexed
%       p_tx(time step, sensor x position), where the sensor spacing is given
%       by dx, the temporal spacing given by dt, and the sound speed in the
%       propagation medium (which is assumed to be acoustically
%       homogeneous) is given by c. The output p_zx is indexed as
%       p_zx(sensor z position, sensor x position).
%
%       The code uses a k-space algorithm which performs (1) a Fourier
%       transform on the data p_tx along both t and x dimensions (into
%       wavenumber-frequency space, where the wavenumber component is along
%       the detector line), (2) a mapping, based on the dispersion relation
%       for a plane wave in an acoustically homogeneous medium, from
%       wavenumber-frequency space to wavenumber-wavenumber space (where
%       the second component is orthogonal to the detector line), and
%       finally (3) an inverse Fourier transform back from the wavenumber
%       domain to the spatial domain. The result is an estimate of the
%       initial acoustic pressure distribution from which the acoustic
%       waves originated. 
%
%       Steps (1) and (3) can be performed efficiently using the fast
%       Fourier transform (FFT); they are therefore fastest when the number
%       of samples and number of detector points are both powers of 2. The
%       mapping in step (2) requires an interpolation of the data from an
%       evenly spaced grid of points in the wavenumber-frequency domain to
%       an evenly-spaced grid of points in the wavenumber-wavenumber
%       domain. The option 'Interp' may be used to choose the interpolation
%       method.
%
%       The physics of photoacoustics requires that the acoustic pressure
%       is initially non-negative everywhere. The estimate of the initial
%       pressure distribution generated by this code may have negative
%       regions due to artefacts arising from differences between the
%       assumed model and the real situation, e.g., homogeneous medium vs.
%       real, somewhat heterogeneous, medium; infinite measurement surface
%       vs. finite-sized region-of-detection, etc. A positivity (or
%       non-negativity) condition can be enforced by setting the optional
%       'PosCond' to true which simply sets any negative parts of the final
%       image to zero. 
%
% USAGE:
%       p_zx = kspaceLineRecon(p_tx, dx, dt, c)
%       p_zx = kspaceLineRecon(p_tx, dx, dt, c, ...)
%
% INPUTS:
%       p_tx        - pressure time-series recorded over an evenly spaced
%                     array of sensor points on a line (indexed as t, x)
%       dx          - spatial step [m]
%       dt          - time step [s]
%       c           - acoustically-homogeneous sound speed [m/s]
%
% OPTIONAL INPUTS:
%       Optional 'string', value pairs that may be used to modify the
%       default computational settings.
%
%       'DataOrder' - String input which sets the data order (default =
%                     'tx'). Valid inputs are 'tx' and 'xt'
%       'Interp'    - string input controlling the interpolation method
%                     used by interp2 in the reconstruction (default =
%                     '*nearest')
%       'PlotRecon' - Boolean controlling whether a plot of the
%                     reconstructed estimate of the initial acoustic
%                     pressure distribution is produced (default = false).
%       'PosCond'   - Boolean controlling whether a positivity condition is
%                     enforced on the reconstructed estimate of the initial
%                     acoustic pressure distribution (default = false).
%
% OUTPUTS:
%       p_zx        - estimate of the initial acoustic pressure
%                     distribution (indexed as z, x)
%
% ABOUT:
%       author      - Bradley Treeby and Ben Cox
%       date        - 11th January 2009
%       last update - 7th February 2011
%       
% This function is part of the k-Wave Toolbox (http://www.k-wave.org)
% Copyright (C) 2009, 2010, 2011 Bradley Treeby and Ben Cox
%
% See also interp2, kspacePlaneRecon, makeGrid

% This file is part of k-Wave. k-Wave is free software: you can
% redistribute it and/or modify it under the terms of the GNU Lesser
% General Public License as published by the Free Software Foundation,
% either version 3 of the License, or (at your option) any later version.
% 
% k-Wave is distributed in the hope that it will be useful, but WITHOUT ANY
% WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
% FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for
% more details. 
% 
% You should have received a copy of the GNU Lesser General Public License
% along with k-Wave. If not, see <http://www.gnu.org/licenses/>.

% start timer
tic;

% define defaults
num_req_inputs = 4;
data_order = 'tx';
interp_method = '*nearest';
plot_recon = false;
positivity_cond = false;

% replace with user defined values if provided
if nargin < num_req_inputs
    error('Incorrect number of inputs');
elseif ~isempty(varargin)
    for input_index = 1:2:length(varargin)
        switch varargin{input_index}
            case 'DataOrder'
                data_order = varargin{input_index + 1};
                if ~strcmp(data_order, 'tx') && ~strcmp(data_order, 'xt')
                    error('Unknown setting for optional input DataOrder');
                end
            case 'Interp'
                interp_method = varargin{input_index + 1};
            case 'PlotRecon'
                plot_recon = varargin{input_index + 1};
            case 'PosCond'
                positivity_cond = varargin{input_index + 1};                
            otherwise
                error('Unknown optional input');
        end
    end
end

% reorder the data if needed (p_tx)
if strcmp(data_order, 'xt')
    p = p.';
end

% mirror the time domain data about t = 0 to allow the cosine transform to
% be computed using an FFT (p_tx)
p = [flipud(p(2:end, :)); p];

% extract the size of mirrored input data
[Nt, Nx] = size(p);

% update command line status
disp('Running k-space line reconstruction...');
disp(['  grid size: ' num2str(Nx) ' by ' num2str((Nt+1)/2) ' pixels']);
disp(['  interpolation mode: ' interp_method]);

% create a computational grid that is evenly spaced in kx and w, where 
% Nz = Nt and dz = dt*c
kgrid = makeGrid(Nx, dx, Nt, dt*c);

% from the grid for kz, create a computational grid for w using the
% relation dz = dt*c; this represents the initial sampling of w 
w = c*kgrid.kz;

% create a second computational grid for w that is based on the real part
% of the dispersion relation w/c = (kx^2 + kz^2)^1/2. This gives an w grid
% that is evenly spaced in kz. This is used for the interpolation from
% p(w, kx) to p(kz, kx). Only real w is taken to force kz (and thus z) to
% be symmetrical about 0 after the interpolation.
w_new = (c*kgrid.k);

% calculate the scaling factor using the value of kz, where
% kz = sqrt( (w/c).^2 - kgrid.kx.^2);
% and then manually replacing the DC value with its limit (otherwise NaN
% results)
sf = c^2*sqrt( (w/c).^2 - kgrid.kx.^2 )./(2*w);
sf(w == 0 & kgrid.kx == 0) = c/2;

% compute the FFT of the input data and scale (p_tx -> p_wkx)
p = sf.*fftshift(fftn(fftshift(p)));

% remove unused variables
clear sf;

% exclude the inhomogeneous part of the wave
p(abs(w) < abs(c*kgrid.kx)) = 0;

% compute the interpolation from p(w, kx) to p(kz, kx) (p_wkx -> p_kzkx)
p = interp2(kgrid.kx, w, p, kgrid.kx, w_new, interp_method);

% set values outside the interpolation range to zero
p(isnan(p)) = 0;

% compute the inverse FFT (p_kzkx -> p_zx)
p = real(ifftshift(ifftn(ifftshift(p))));

% remove the left part of the mirrored data which corresponds to the
% negative part of the mirrored time data
p = p( (Nt + 1)/2:Nt, :);

% correct the scaling - the forward FFT is computed with a spacing of dt
% and the reverse requires a spacing of dz = dt*c, the reconstruction
% assumes that p0 is symmetrical about z, and only half the plane collects
% data (first approximation to correcting the limited view problem)
p = 2*2*p./c;

% enfore positivity condition
if positivity_cond
    disp('  applying positivity condition...');
    p(p < 0) = 0;
end

% update command line status
disp(['  computation completed in ' scaleTime(toc)]);

% plot the reconstruction
if plot_recon
    
    % allocate axis dimensions
    x_axis = [0 Nx*dx];
    z_axis = [0 (Nt/2)*dt*c]; 
    
    % select suitable axis scaling factor
    [x_sc scale prefix] = scaleSI(max([Nx*dx, (Nt/2)*dt*c])); 
    
    % select suitable plot scaling factor
    plot_scale = max(p(:));
    
    % create the figure
    figure;
    imagesc(x_axis*scale, z_axis*scale, p, [-plot_scale, plot_scale]);
    axis image;
    colormap(getColorMap);
    xlabel(['Sensor Position [' prefix 'm]']);
    ylabel(['Depth [' prefix 'm]']);
    colorbar;
end